#ifndef KCAS_H
#define KCAS_H

#include "../descriptors/descriptors_impl2.h"

#define kcastagptr_t uintptr_t
#define kcasptr_t kcasdesc_t<K,NPROC>*
#define rdcsstagptr_t uintptr_t
#define rdcssptr_t rdcssdesc_t*
#define casword_t uintptr_t

#define KCAS_STATE_UNDECIDED 0
#define KCAS_STATE_SUCCEEDED 4
#define KCAS_STATE_FAILED 8

#define KCAS_LEFTSHIFT 2

struct rdcssdesc_t {
    volatile seqbits_t seqBits;
    casword_t volatile * addr1;
    casword_t old1;
    casword_t volatile * addr2;
    casword_t old2;
    casword_t new2;
    const static int size = sizeof(seqBits)+sizeof(addr1)+sizeof(old1)+sizeof(addr2)+sizeof(old2)+sizeof(new2);
    volatile char padding[128+((64-size%64)%64)]; // add padding to prevent false sharing
};

struct kcasentry_t { // just part of kcasdesc_t, not a standalone descriptor
    casword_t volatile * addr;
    casword_t oldval;
    casword_t newval;
};

template <int K, int NPROC>
class kcasdesc_t {
public:
    volatile seqbits_t seqBits;
    casword_t numEntries;
    kcasentry_t entries[K];
    const static int size = sizeof(seqBits)+sizeof(numEntries)+sizeof(entries);
    volatile char padding[128+((64-size%64)%64)]; // add padding to prevent false sharing
    
    void addValAddr(casword_t * addr, casword_t oldval, casword_t newval) {
        entries[numEntries].addr = addr;
        entries[numEntries].oldval = oldval << KCAS_LEFTSHIFT;
        entries[numEntries].newval = newval << KCAS_LEFTSHIFT;
        ++numEntries;
        //std::cout<<"addr="<<(size_t) addr<<" oldval="<<oldval<<" newval="<<newval<<" numEntries="<<std::endl;
    }
    
    void addPtrAddr(casword_t * addr, casword_t oldval, casword_t newval) {
        entries[numEntries].addr = addr;
        entries[numEntries].oldval = oldval;
        entries[numEntries].newval = newval;
        ++numEntries;
    }
};

template <int K, int NPROC>
class kcasProvider {
    /**
     * Data definitions
     */
private:
    // descriptor reduction algorithm
    #define KCAS_SEQBITS_OFFSET_STATE 0
    #define KCAS_SEQBITS_MASK_STATE 0xf
    #define KCAS_SEQBITS_NEW(seqBits) \
        ((((seqBits)&MASK_SEQ)+(1<<OFFSET_SEQ)) \
        | (KCAS_STATE_UNDECIDED<<KCAS_SEQBITS_OFFSET_STATE))
    #define RDCSS_SEQBITS_NEW(seqBits) \
        (((seqBits)&MASK_SEQ)+(1<<OFFSET_SEQ))
    volatile char __padding_desc[128];
    kcasdesc_t<K,NPROC> kcasDescriptors[LAST_TID+1] __attribute__ ((aligned(64)));
    rdcssdesc_t rdcssDescriptors[LAST_TID+1] __attribute__ ((aligned(64)));
    volatile char __padding_desc3[128];

    /**
     * Function declarations
     */
public:
    kcasProvider();
    void writePtr(casword_t volatile * addr, casword_t val);
    void writeVal(casword_t volatile * addr, casword_t val);
    casword_t readPtr(const int tid, casword_t volatile * addr);
    casword_t readVal(const int tid, casword_t volatile * addr);
    int kcas(const int tid, kcasptr_t ptr);
    kcasptr_t createKcasDesc(const int tid);
private:
    bool help(const int tid, kcastagptr_t tagptr, kcasptr_t ptr, bool helpingOther);
    void helpOther(const int tid, kcastagptr_t tagptr);
    casword_t rdcssRead(const int tid, casword_t volatile * addr);
    casword_t rdcss(const int tid, rdcssptr_t ptr, rdcsstagptr_t tagptr);
    void rdcssHelp(rdcsstagptr_t tagptr, rdcssptr_t snapshot, bool helpingOther);
    void rdcssHelpOther(rdcsstagptr_t tagptr);
    std::string tagptrToString(casword_t tagptr);
};

#endif