/* 
 * File:   test_kcas.cpp
 * Author: tabrown
 *
 * Created on June 23, 2016, 5:01 PM
 */

#include <iostream>
#include "kcas_reuse.h"
using namespace std;

#define KCAS_MAXK 8
#define KCAS_MAXTHREADS 64

int main(int argc, char** argv) {
    const int tid = 0; // some kcas provider functions expect a thread id in 0 ... KCAS_MAXTHREADS-1

    // initialize kcas provider
    kcasProvider<KCAS_MAXK, KCAS_MAXTHREADS> prov;
    
    // initialize simple array data structure to perform kcas operations on
    casword_t arr[100];
    for (int i=0;i<100;++i) {
        prov.writeVal(&arr[i], 0);
    }

    /**
     * example kcas: increment each slot in arr[0..2] by 1
     */
    
    // step 1: create a kcas descriptor
    kcasdesc_t<KCAS_MAXK, KCAS_MAXTHREADS> * desc = prov.createKcasDesc(tid);
    
    // step 2: specify (addr, old, new) triples for the kcas
    for (int i=0;i<3;++i) {
        // note: you must use prov.readVal or prov.readPtr to read any data that
        //       might be changed by a kcas. (regular reads might not be correct.)
        //       use readPtr when the address contains a pointer
        //       (meaning that the least significant bit is unused,
        //        and can be stolen by the kcas algorithm),
        //       and readVal when the address contains a value
        //       that might use the least significant bit.
        //       the kcas algorithm will automatically shift the value left so
        //       it can steal the LSB (in fact, we steal two LSBs).
        //       note that this means you can't store values that need all the
        //       bits of a machine word. you must leave 2 bits for the kcas alg.
        casword_t oldval = prov.readVal(tid, &arr[i]);
        // to add an (addr, old, new) triple to the kcas descriptor:
        //       use addPtrAddr when the address contains a pointer,
        //       and addValAddr when the address contains a value.
        desc->addValAddr(&arr[i], oldval, oldval+1);
    }
    
    // step 2: invoke kcas, passing in the descriptor pointer
    int result = prov.kcas(tid, desc);
    
    cout<<"result="<<result<<" arr[0]="<<prov.readVal(tid, &arr[0])<<" arr[1]="<<prov.readVal(tid, &arr[1])<<" arr[2]="<<prov.readVal(tid, &arr[2])<<" arr[3]="<<prov.readVal(tid, &arr[3])<<endl;
    
    return 0;
}

